from com.arm.debug.dtsl.configurations import DTSLv1
from com.arm.debug.dtsl.components import FormatterMode
from com.arm.debug.dtsl.components import CSDAP
from com.arm.debug.dtsl.components import MemoryRouter
from com.arm.debug.dtsl.components import DapMemoryAccessor
from com.arm.debug.dtsl.components import Device
from com.arm.debug.dtsl.components import DSTREAMTraceCapture
from com.arm.debug.dtsl.components import V7M_CSTPIU
from com.arm.debug.dtsl.components import ITMTraceSource

NUM_CORES_CORTEX_M3 = 1
ATB_ID_BASE = 2
DSTREAM_PORTWIDTH = 4
ITM_ATB_ID = 1
ITM_FUNNEL_PORT = 3


class DtslScript(DTSLv1):
    @staticmethod
    def getOptionList():
        return [
            DTSLv1.tabSet("options", "Options", childOptions=[
                DTSLv1.tabPage("trace", "Trace Capture", childOptions=[
                    DTSLv1.enumOption('traceCapture', 'Trace capture method', defaultValue="none",
                        values = [("none", "None"), ("DSTREAM", "DSTREAM 4GB Trace Buffer")],
                        setter=DtslScript.setTraceCaptureMethod),
                ]),
                DTSLv1.tabPage("itm", "ITM", childOptions=[
                    DTSLv1.booleanOption('itm', 'Enable ITM trace', defaultValue=False,
                        setter=DtslScript.setITMEnabled),
                ]),
            ])
        ]
    
    def __init__(self, root):
        DTSLv1.__init__(self, root)
        
        # locate devices on the platform and create corresponding objects
        self.discoverDevices()
        
        self.mgdPlatformDevs = set()
        
        # tracks which devices are managed when a trace mode is enabled
        self.mgdTraceDevs = {}
        
        # only DAP device is managed by default - others will be added when enabling trace, SMP etc
        self.mgdPlatformDevs.add(self.dap)
        
        self.exposeCores()
        
        self.setupDSTREAMTrace(DSTREAM_PORTWIDTH)
        
        self.setManagedDevices(self.mgdPlatformDevs)
    
    # +----------------------------+
    # | Target dependent functions |
    # +----------------------------+
    
    def discoverDevices(self):
        '''find and create devices'''
        
        dapDev = self.findDevice("ARMCS-DP")
        self.dap = CSDAP(self, dapDev, "DAP")
        
        cortexM3coreDev = 0
        self.cortexM3cores = []
        
        streamID = ATB_ID_BASE
        
        for i in range(0, NUM_CORES_CORTEX_M3):
            # create core
            cortexM3coreDev = self.findDevice("Cortex-M3", cortexM3coreDev+1)
            dev = Device(self, cortexM3coreDev, "Cortex-M3")
            self.cortexM3cores.append(dev)
            
        # DSTREAM
        self.DSTREAM = DSTREAMTraceCapture(self, "DSTREAM")
        
        # TPIU
        tpiuDev = self.findDevice("CSTPIU")
        self.tpiu = self.createTPIU(tpiuDev, "TPIU")
        # ITM
        itmDev = self.findDevice("CSITM")
        self.ITM = self.createITM(itmDev, ITM_ATB_ID, "ITM")
        
    def exposeCores(self):
        for core in self.cortexM3cores:
            self.addDeviceInterface(self.createDAPWrapper(core))
    
    def setupDSTREAMTrace(self, portwidth):
        '''Setup DSTREAM trace capture'''
        
        self.tpiu.setPortSize(portwidth)
        
        # configure the DSTREAM for continuous trace
        self.DSTREAM.setTraceMode(DSTREAMTraceCapture.TraceMode.Continuous)
        self.DSTREAM.setPortWidth(portwidth)
        
        # register other trace components
        self.DSTREAM.setTraceComponentOrder([ self.tpiu ])
        
        # register the DSTREAM with the configuration
        self.addTraceCaptureInterface(self.DSTREAM)
        
        # automatically handle connection/disconnection to trace components
        self.addManagedTraceDevices("DSTREAM", [ self.tpiu, self.DSTREAM ])
        
        # register trace sources
        self.registerTraceSources(self.DSTREAM)
    
    def setTraceSourceEnabled(self, source, enabled):
        '''Enable/disable a trace source'''
        source.setEnabled(enabled)
    
    def createTPIU(self, tpiuDev, name):
        tpiu = V7M_CSTPIU(self, tpiuDev, name, self.cortexM3cores[0])
        # disable the TPIU by default to allow ETB to work at full rate
        tpiu.setEnabled(False)
        return tpiu
    
    def setDSTREAMTraceEnabled(self, enabled):
        '''Enable/disable DSTREAM trace capture'''
        self.tpiu.setEnabled(enabled)
    
    def registerTraceSources(self, traceCapture):
        '''Register all trace sources with trace capture device'''
        self.registerTraceSource(traceCapture, self.ITM)
    
    def getFunnelPortForSource(self, source):
        '''Get the funnel port number for a trace source'''
        
        # Build map of sources to funnel ports
        portMap = {self.ITM: ITM_FUNNEL_PORT}
        
        return portMap.get(source, None)
    
    # +--------------------------------+
    # | Callback functions for options |
    # +--------------------------------+
    
    def optionValuesChanged(self):
        '''Callback to update the configuration state after options are changed'''
        optionValues = self.getOptionValues()
        traceMode = optionValues.get("options.trace.traceCapture")
        self.setManagedDevices(self.getManagedDevices(traceMode))
        
    def getManagedDevices(self, traceKey):
        '''Get the required set of managed devices for this configuration'''
        return self.mgdPlatformDevs | self.mgdTraceDevs.get(traceKey, set())
    
    def setTraceCaptureMethod(self, method):
        if method == "none":
            self.setDSTREAMTraceEnabled(False)
        elif method == "DSTREAM":
            self.setDSTREAMTraceEnabled(True)
    
    def setITMEnabled(self, enabled):
        '''Enable/disable the ITM trace source'''
        self.setTraceSourceEnabled(self.ITM, enabled)
    
    # +------------------------------+
    # | Target independent functions |
    # +------------------------------+
    
    def addManagedPlatformDevices(self, devs):
        '''Add devices to the set of devices managed by the configuration'''
        for d in devs:
            self.mgdPlatformDevs.add(d)
    
    def registerTraceSource(self, traceCapture, source):
        '''Register trace source with trace capture device'''
        traceCapture.addTraceSource(source)
        self.addManagedTraceDevices(traceCapture.getName(), [ source ])
    
    def addManagedTraceDevices(self, traceKey, devs):
        '''Add devices to the set of devices managed by the configuration for this trace mode'''
        traceDevs = self.mgdTraceDevs.get(traceKey)
        if not traceDevs:
            traceDevs = set()
            self.mgdTraceDevs[traceKey] = traceDevs
        for d in devs:
            traceDevs.add(d)
    
    def createDAPWrapper(self, core):
        '''Add a wrapper around a core to allow access to AHB and APB via the DAP'''
        return MemoryRouter(
            [DapMemoryAccessor("AHB", self.dap, 0, "AHB bus accessed via AP_0 on DAP_0"),
             DapMemoryAccessor("APB", self.dap, 1, "APB bus accessed via AP_1 on DAP_0")],
            core)
    
    def createITM(self, itmDev, streamID, name):
        itm = ITMTraceSource(self, itmDev, streamID, name)
        # disabled by default - will enable with option
        itm.setEnabled(False)
        return itm
    

